open Edecl;;
open Ertnl;;

let eatspaces s =
 let j = ref 0 in
 for i = 0 to String.length s - 1 do
  if s.[i] <> ' ' then (s.[!j] <- s.[i]; incr j)
 done;
 String.sub s 0 !j;;

let int_of_ascii x = int_of_char x - 48;;
let apost = char_of_int 39;;

exception Parse of string;;

let parse s =
 let s = (eatspaces (String.copy s)) ^ ";" in
 let i = ref 0 in
 let parse_num () =
  let rec pn acc =
   match s.[!i] with
     '0'..'9' as x -> incr i; pn (10 * acc + (int_of_ascii x))
    |_ -> Num acc in pn 0 in
 let parse_var () =
  let rec pn acc =
   match s.[!i] with
     'a'..'z' as x -> incr i; pn (acc ^ (String.make 1 x))
    |'0'..'9' as x -> incr i; pn (acc ^ (String.make 1 x))
    |x when x = apost -> incr i; pn (acc ^ "'")
    |_ -> Var acc in pn "" in
 let rec parse_add () =
  let rec pp acc =
   match s.[!i] with
    '+' -> incr i; pp (Add (acc, parse_mul ()))
   |'-' -> incr i; pp (Add (acc, Minus (parse_mul ())))
   | _  -> acc in pp (parse_mul ())
 and parse_mul () =
  let rec pp acc =
   match s.[!i] with
    '*' -> incr i; pp (Mul (acc, parse_pow ()))
   |'/' -> incr i; pp (Frac(acc, parse_pow ()))
   | _  -> acc in pp (parse_pow ())
 and parse_pow () =
  let rec pp acc =
   match s.[!i] with
    '^' -> incr i; pp (Pow (acc, parse_brackets ()))
   |'@' -> incr i; pp (Root(acc, parse_brackets ()))
   | _  -> acc in pp (parse_brackets ())
 and parse_brackets () =
  match s.[!i] with
    '0'..'9' -> parse_num ()
   |'a'..'z' -> parse_var ()
   |'+' -> incr i; parse_brackets ()
   |'-' -> incr i; Minus (parse_brackets ())
   |'(' -> incr i; let t = parse_add () in
           if s.[!i] <> ')' then raise (Parse ("missing bracket at " ^ (string_of_int !i)))
            else (incr i; t)
   |';' -> raise (Parse "end reached unexpectedly")
   | _  -> raise (Parse ("unexpected symbol at " ^ (string_of_int !i))) in
 let res = parse_add () in
 if s.[!i] <> ';' then raise (Parse ("unexpected symbol at " ^ (string_of_int !i))) else res;;


let string_of_tree expr =
 let rec sot t pr =
  let br i = if pr < i then ("", "") else ("(", ")") in
  match t with
    Num x -> string_of_int x
   |Var x -> x
   |Minus x -> (match x with Num _ | Var _ -> "(-" | _ -> " - (") ^ (sot x 0) ^ ")"
   |Add (a, Minus b) -> let (lb, rb) = br 1 in lb ^ (sot a 0) ^ " - " ^ (sot b 1) ^ rb
   |Add (Add (a, b), c) -> let (lb, rb) = br 1 in lb ^ (sot (Add (a, b)) 0) ^ " + " ^ (sot c 1) ^ rb
   |Mul (Mul (a, b), c) -> let (lb, rb) = br 2 in lb ^ (sot (Mul (a, b)) 1) ^ " * " ^ (sot c 2) ^ rb
   |Frac(Frac(a, b), c) -> let (lb, rb) = br 2 in lb ^ (sot (Frac(a, b)) 1) ^ " / " ^ (sot c 2) ^ rb
   |Pow (Pow (a, b), c) -> let (lb, rb) = br 3 in lb ^ (sot (Pow (a, b)) 2) ^  "^"  ^ (sot c 3) ^ rb
   |Root(Root(a, b), c) -> let (lb, rb) = br 3 in lb ^ (sot (Root(a, b)) 2) ^  "@"  ^ (sot c 3) ^ rb
   |Add (a, Num x) when x < 0 -> let (lb, rb) = br 2 in lb ^ (sot a 0) ^ "-" ^ (string_of_int (-x)) ^ rb
   |Add (a, b) -> let (lb, rb) = br 1 in lb ^ (sot a 0) ^ " + " ^ (sot b 1) ^ rb
   |Mul (a, b) -> let (lb, rb) = br 2 in lb ^ (sot a 1) ^ " * " ^ (sot b 2) ^ rb
   |Frac(a, b) -> let (lb, rb) = br 2 in lb ^ (sot a 1) ^ " / " ^ (sot b 2) ^ rb
   |Pow (a, b) -> let (lb, rb) = br 3 in lb ^ (sot a 2) ^  "^"  ^ (sot b 3) ^ rb
   |Root(a, b) -> let (lb, rb) = br 3 in lb ^ (sot a 2) ^  "@"  ^ (sot b 3) ^ rb
 in sot expr 0;;

let rec string_of_tree2 expr =
 match expr with
   Num x -> string_of_int x
  |Var x -> x
  |Minus x -> " - (" ^ (string_of_tree x) ^ ")"
  |Add (a, b) -> "(" ^ (string_of_tree a) ^ " + " ^ (string_of_tree b) ^ ")"
  |Mul (a, b) -> "(" ^ (string_of_tree a) ^ " * " ^ (string_of_tree b) ^ ")"
  |Frac(a, b) -> "(" ^ (string_of_tree a) ^ " / " ^ (string_of_tree b) ^ ")"
  |Pow (a, b) -> "(" ^ (string_of_tree a) ^  "^"  ^ (string_of_tree b) ^ ")"
  |Root(a, b) -> "(" ^ (string_of_tree a) ^  "@"  ^ (string_of_tree b) ^ ")";;

let string_of_rational (x, y) =
 if y = 1 then string_of_int x else
 "(" ^ (string_of_int x) ^ "/" ^ (string_of_int y) ^ ")";;

let string_of_poly expr =
 let split_vars vars =
  let rec split_vars vars (pos, neg) =
  match vars with
    (name, (pn, pd))::ls when pn < 0 -> split_vars ls (pos, (name, (-pn, pd))::neg)
   |l1::ls -> split_vars ls (l1::pos, neg)
   |[] -> (List.rev pos, List.rev neg) in split_vars vars ([], []) in
 let proceed_vars vars = (List.fold_left (fun acc (name, power) -> acc ^
                           (if power = (1, 1) then name else
                                "(" ^ name ^ "^" ^ (string_of_rational power) ^ ")")) "" vars) in
 let conv_sign x = if x = 1 then "" else if x = -1 then "-" else string_of_int x in
 let string_of_mul (factor, vars) =
  let (pos, neg) = split_vars vars in
    match (proceed_vars pos, proceed_vars neg) with
      ("" , "" ) -> string_of_rational factor
     |(pos, "" ) -> (conv_sign (fst factor)) ^ pos ^ (if snd factor = 1 then "" else "/" ^ (string_of_int (snd factor)))
     |("" , neg) -> let (lb, ffac) = let ffac = fst factor in if ffac < 0 then ("-(", -ffac) else ("(", ffac) in
                    lb ^ (string_of_int ffac) ^ "/" ^ (conv_sign (snd factor)) ^ neg ^ ")"
     |(pos, neg) -> let (lb, ffac) = let ffac = fst factor in if ffac < 0 then ("-(", -ffac) else ("(", ffac) in
                    let (ffac, denom) = (conv_sign (fst factor), conv_sign (snd factor)) in
                    lb ^ ffac ^ pos ^ "/" ^ denom ^ neg ^ ")" in
 let rec string_of_poly expr acc =
  match expr with
    l1::[] -> acc ^ (string_of_mul l1)
   |l1::l2::ls -> string_of_poly (l2::ls) (acc ^ (string_of_mul l1) ^ (if comp_rationals (0, 1) (fst l2) then "" else "+"))
   |[] -> "" in
 string_of_poly expr "";;